#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <status.h>
#include <elflib.h>

#ifdef HAVE_PUBLIB_H
#include <publib.h>
#else
#define xmalloc(s) malloc(s)
#define xfree(p) free(p)
#endif

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#define assume(expr) \
    if (!(expr)) { \
    fprintf(stderr, "Assumption %s at %s:%d is false\n", __STRING(expr), \
    __FILE__, __LINE__); abort(); }

Status
Elf_EhdrRead(
    FILE	*f,
    Elf32_Ehdr	*ehdrPtr)
{
    int		len;
    int		n;
    Status	status;

    len = sizeof(Elf32_Ehdr);
    n = fread(ehdrPtr, len, 1, f);
    if (n != 1) {
	perror("Unable to read Elf header");
	status = FAILURE;
	goto done;
    }
    if (ehdrPtr->e_ehsize != sizeof(Elf32_Ehdr)) {
	status = ELF_HDR_BAD_SIZE;
	goto done;
    }
    if ((ehdrPtr->e_ident[EI_MAG0] != 0x7f) ||
	(ehdrPtr->e_ident[EI_MAG1] != 'E') ||
	(ehdrPtr->e_ident[EI_MAG2] != 'L') ||
	(ehdrPtr->e_ident[EI_MAG3] != 'F')) {
	status = ELF_HDR_BAD_MAGIC;
	goto done;
    }
    if (ehdrPtr->e_ident[EI_CLASS] != ELFCLASS32) {
	status = ELF_HDR_BAD_CLASS;
	goto done;
    }
    if (ehdrPtr->e_ident[EI_DATA] != ELFDATA2LSB) {
	status = ELF_HDR_BAD_DATA;
	goto done;
    }
    if (ehdrPtr->e_ident[EI_VERSION] != EV_CURRENT) {
	status = ELF_HDR_BAD_VERSION;
	goto done;
    }
    if (ehdrPtr->e_shentsize != sizeof(Elf32_Shdr)) {
	status = ELF_SHDR_BAD_SIZE;
	goto done;
    }
    if (ehdrPtr->e_phentsize != sizeof(Elf32_Phdr)) {
	status = ELF_PHDR_BAD_SIZE;
	goto done;
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_PhdrRead(
    FILE	*f,
    Elf32_Ehdr	*ehdrPtr,
    int		index,
    Elf32_Phdr	*phdrPtr)
{
    Status 	status;
    int		n;

    if ((index < 0) || (index >= ehdrPtr->e_phnum)) {
	status = ELF_OUT_OF_BOUNDS;
	goto done;
    }
    n = fseek(f, ehdrPtr->e_phoff + (index * sizeof(Elf32_Phdr)), 
	    SEEK_SET);
    if (n != 0) {
	status = ELF_IO_ERROR;
	goto done;
    }
    n = fread(phdrPtr, sizeof(Elf32_Phdr), 1, f);
    if (n != 1) {
	status = ELF_IO_ERROR;
	goto done;
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_PhdrReadAll(
    FILE	*f,
    Elf32_Ehdr	*ehdrPtr,
    Elf32_Phdr	*phdrs)
{
    int		n;
    Status	status;
    int		i;

    n = fseek(f, ehdrPtr->e_phoff, SEEK_SET);
    if (n != 0) {
	status = ELF_IO_ERROR;
	goto done;
    }
    for (i = 0; i < ehdrPtr->e_phnum; i++) {
	n = fread(&phdrs[i], sizeof(Elf32_Phdr), 1, f);
	if (n != 1) {
	    status = ELF_IO_ERROR;
	    goto done;
	}
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_ShdrRead(
    FILE	*f,
    Elf32_Ehdr	*ehdrPtr,
    int		index,
    Elf32_Shdr	*shdrPtr)
{
    Status	status;
    int		n;

    if ((index < 0) || (index >= ehdrPtr->e_shnum)) {
	status = ELF_OUT_OF_BOUNDS;
	goto done;
    }
    n = fseek(f, ehdrPtr->e_shoff + (index * sizeof(Elf32_Shdr)), 
	    SEEK_SET);
    if (n != 0) {
	status = ELF_IO_ERROR;
	goto done;
    }
    n = fread(shdrPtr, sizeof(Elf32_Shdr), 1, f);
    if (n != 1) {
	status = ELF_IO_ERROR;
	goto done;
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_ShdrReadAll(
    FILE	*f,
    Elf32_Ehdr	*ehdrPtr,
    Elf32_Shdr	*shdrs)
{
    int		n;
    Status	status;
    int		i;

    n = fseek(f, ehdrPtr->e_shoff, SEEK_SET);
    if (n != 0) {
	status = ELF_IO_ERROR;
	goto done;
    }
    for (i = 0; i < ehdrPtr->e_shnum; i++) {
	n = fread(&shdrs[i], sizeof(Elf32_Shdr), 1, f);
	if (n != 1) {
	    status = ELF_IO_ERROR;
	    goto done;
	}
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_SectionRead(
    FILE	*f,
    Elf32_Shdr	*shdrPtr,
    void	**sectionPtr)   // OUT
{
    Status	status;
    int		len;
    void	*section;
    int		n;

    len = shdrPtr->sh_size;
    if ((shdrPtr->sh_type != SHT_NOBITS) && 
	(shdrPtr->sh_type != SHT_NULL) && (len != 0)) {
	section = xmalloc(len);
	n = fseek(f, shdrPtr->sh_offset, SEEK_SET);
	if (n != 0) {
	    status = ELF_IO_ERROR;
	    goto done;
	}
	n = fread(section, len, 1, f);
	if (n != 1) {
	    status = ELF_IO_ERROR;
	    goto done;
	}
    } else {
	section = NULL;
    }
    status = SUCCESS;
    *sectionPtr = section;
done:
    return status;
}

Status
Elf_SectionWrite(
    FILE	*f,
    Elf32_Shdr	*shdrPtr,
    void	*section)
{
    Status	status;
    int		len;
    int		n;

    len = shdrPtr->sh_size;
    if ((shdrPtr->sh_type != SHT_NOBITS) && 
	(shdrPtr->sh_type != SHT_NULL) && (len != 0)) {
	n = fseek(f, shdrPtr->sh_offset, SEEK_SET);
	if (n != 0) {
	    status = ELF_IO_ERROR;
	    goto done;
	}
	n = fwrite(section, len, 1, f);
	if (n != 1) {
	    status = ELF_IO_ERROR;
	    goto done;
	}
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_SectionGetName(
    FILE	*f,
    Elf32_Ehdr	*ehdrPtr,
    Elf32_Shdr	*shdrPtr,
    char	**namePtr)  // OUT
{
    Status	status;
    static FILE	*lastFile = NULL;
    static char	*names = NULL;
    static int	nameLen;
    int		index;

    if (f != lastFile) {
	Elf32_Shdr	strtabShdr;

	status = Elf_ShdrRead(f, ehdrPtr, ehdrPtr->e_shstrndx, &strtabShdr);
	if (status != SUCCESS) {
	    goto done;
	}
	if (names != NULL) {
	    xfree(names);
	    names = NULL;
	}
	status = Elf_SectionRead(f, &strtabShdr, (void **) &names);
	if (status != SUCCESS) {
	    goto done;
	}
	nameLen = strtabShdr.sh_size;
	lastFile = f;
    }
    index = shdrPtr->sh_name;
    assume((index >= 0) && (index < nameLen));
    *namePtr = strdup(&names[index]);
    status = SUCCESS;
done:
    return status;
}

Status
Elf_FileRead(
    FILE	*f,
    Elf_File	*filePtr)
{
    Status	status = FAILURE;
    Elf_Section	*sectionPtr;
    Elf32_Shdr	*shdrPtr;
    int		i;
    Status	rs;
    Elf_Section	*stringTablePtr;
    char	*names;
    List_Links	*itemPtr;
    int		size;
    static int 	id = 0;
    Elf_Section	*prevPtr;

    filePtr->id = id++;
    status = Elf_EhdrRead(f, &filePtr->ehdr);
    if (status != SUCCESS) {
	goto done;
    }
    filePtr->hdrSize = filePtr->ehdr.e_ehsize;
    /*
     * Read the segments.
     */
    filePtr->numSegments = filePtr->ehdr.e_phnum;
    filePtr->hdrSize += filePtr->numSegments * sizeof(Elf32_Phdr);
#ifdef DEBUG
    printf("File %s has %d bytes of headers\n", filePtr->name, 
	    filePtr->hdrSize);
#endif /* DEBUG */
    for (i = 0; i < filePtr->numSegments; i++) {
	Elf_Segment 	*segmentPtr;

	segmentPtr = xmalloc(sizeof(Elf_Segment));
	status = Elf_PhdrRead(f, &filePtr->ehdr, i, &segmentPtr->phdr);
	if (status != SUCCESS) {
	    goto done;
	}
	segmentPtr->filename = filePtr->name;
	List_Init(&segmentPtr->links);
	List_Insert(&segmentPtr->links, LIST_ATREAR(&filePtr->segmentList));
    }
    /*
     * Read the section headers.
     */
    filePtr->numSections = filePtr->ehdr.e_shnum;
    prevPtr = NULL;
    for (i = 0; i < filePtr->numSections; i++) {
	sectionPtr = xmalloc(sizeof(Elf_Section));
	status = Elf_ShdrRead(f, &filePtr->ehdr, i, &sectionPtr->shdr);
	/*
	 * For some reason there will sometimes be a gap between sections
	 * that isn't included in the size of the first section. Compute
	 * the real size of the section from the difference in offsets.
	 */
	if (prevPtr != NULL) {
	    int		offset;
	    int		mod;
	    offset = prevPtr->shdr.sh_offset + prevPtr->shdr.sh_size;
	    if (sectionPtr->shdr.sh_addralign > 1) {
		mod = offset % sectionPtr->shdr.sh_addralign;
		if (mod != 0) {
		    offset += sectionPtr->shdr.sh_addralign - mod;
		}
	    }
#ifdef DEBUG
	    if (offset != sectionPtr->shdr.sh_offset) {
		printf("WARNING: section %d isn't contiguous.\n", i);
	    }
#endif /* DEBUG */
	    prevPtr->size = sectionPtr->shdr.sh_offset - 
		prevPtr->shdr.sh_offset;
	}
	sectionPtr->name = NULL;
	sectionPtr->section = NULL;
	sectionPtr->index = i;
	sectionPtr->id.file = filePtr->id;
	sectionPtr->id.index = sectionPtr->index;
	sectionPtr->size = sectionPtr->shdr.sh_size;
	sectionPtr->filename = filePtr->name;
	List_InitElement(&sectionPtr->links);
	List_Insert(&sectionPtr->links, LIST_ATREAR(&filePtr->sectionList));
	if ((sectionPtr->shdr.sh_type != SHT_NULL) &&
	    (sectionPtr->shdr.sh_type != SHT_NOBITS)) {
	    prevPtr = sectionPtr;
	}
    }
    /*
     * Read the sections
     */
    sectionPtr = NULL;
    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	void	*section;

	sectionPtr = (Elf_Section *) itemPtr;
	shdrPtr = &sectionPtr->shdr;
	status = Elf_SectionRead(f, shdrPtr, & section);
	if (status != SUCCESS) {
	    goto done;
	}
	sectionPtr->section = section;
    }
    /* 
     * Create the name strings
     */
    if (filePtr->ehdr.e_shstrndx != SHN_UNDEF) {
	rs = Elf_GetSectionByIndex(filePtr, filePtr->ehdr.e_shstrndx, 
		    &stringTablePtr);
	if (rs != SUCCESS) {
	    status = rs;
	    goto done;
	}
	names = (char *) stringTablePtr->section;
	size = stringTablePtr->shdr.sh_size;
	LIST_FORALL(&filePtr->sectionList, itemPtr) {
	    sectionPtr = (Elf_Section *) itemPtr; 
	    if (sectionPtr->shdr.sh_name >= size) {
		fprintf(stderr, "String index out of bounds.\n");
		abort();
	    }
	    sectionPtr->name = &names[sectionPtr->shdr.sh_name];
	}
    }
    /*
     * Set up the symbol table.
     */
    rs = Elf_GetSectionByName(filePtr, ".dynsym", &filePtr->symbolTablePtr);
    if (rs != SUCCESS) {
	status = rs;
	goto done;
    }
    rs = Elf_GetSectionByName(filePtr, ".dynstr", &filePtr->stringPtr);
    if (rs != SUCCESS) {
	status = rs;
	goto done;
    }
    status = SUCCESS;
done:
    return status;
}

Status
Elf_GetSectionByIndex(
    Elf_File	*filePtr,
    int		index,
    Elf_Section	**sectionPtrPtr)
{
    Status	status;
    int		i;
    Elf_Section	*sectionPtr = NULL;
    List_Links	*itemPtr;

    i = 0;
    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	if (i == index) {
	    sectionPtr = (Elf_Section *) itemPtr; 
	}
	i++;
    }
    if (sectionPtr == NULL) { 
	status = FAILURE;
    } else {
	*sectionPtrPtr = sectionPtr;
	status = SUCCESS;
    }
    return status;
}

Status
Elf_GetSectionByName(
    Elf_File	*filePtr,
    char	*name,
    Elf_Section	**sectionPtrPtr)
{
    Status	status = FAILURE;
    List_Links	*itemPtr;
    Elf_Section	*sectionPtr;

    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	sectionPtr = (Elf_Section *) itemPtr; 
	if (strcmp(sectionPtr->name, name) == 0) {
	    *sectionPtrPtr = sectionPtr;
	    status = SUCCESS;
	    goto done;
	}
    }
    status = FAILURE;
done:
    return status;
}

Status
Elf_GetSectionByAddress(
    Elf_File	*filePtr,
    Elf32_Addr	address,
    Elf_Section	**sectionPtrPtr)
{
    Status	status = FAILURE;
    List_Links	*itemPtr;
    Elf_Section	*sectionPtr;

    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	sectionPtr = (Elf_Section *) itemPtr; 
	if (sectionPtr->shdr.sh_addr == address) {
	    *sectionPtrPtr = sectionPtr;
	    status = SUCCESS;
	    goto done;
	}
    }
    status = FAILURE;
done:
    return status;
}

Status
Elf_GetSectionByID(
    Elf_File	*filePtr,
    int		file,
    int		index,
    Elf_Section	**sectionPtrPtr)
{
    Status	status = FAILURE;
    Elf_Section	*sectionPtr;
    List_Links	*itemPtr;

    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	sectionPtr = (Elf_Section *) itemPtr;
	if ((sectionPtr->id.file == file) && (sectionPtr->id.index == index)) {
	    *sectionPtrPtr = sectionPtr;
	    status = SUCCESS;
	    goto done;
	}
    }
    status = FAILURE;
done:
    return status;
}



Status
Elf_GetSymbol(
    Elf_File		*filePtr,
    char		*name,
    Elf32_Sym		**symPtrPtr)
{
    Status	status = FAILURE;
    Elf_Section	*tablePtr;
    char	*strings;
    Elf32_Sym	*symPtr;
    int		count;
    int		i;

    tablePtr = filePtr->symbolTablePtr;
    strings = (char *) filePtr->stringPtr->section;
    count = tablePtr->shdr.sh_size / sizeof(Elf32_Sym);
    for (i = 0; i < count; i++) {
	symPtr = ((Elf32_Sym *) tablePtr->section) + i;
	if (strcmp(name, &strings[symPtr->st_name]) == 0) {
	    *symPtrPtr = symPtr;
	    status = SUCCESS;
	    goto done;
	}
    }
done:
    return status;
}

Status
Elf_FileWrite(
    FILE	*f,
    Elf_File	*filePtr)
{
    int		n;
    int		i;
    Elf_Section	*sectionPtr;
    int		len;
    List_Links	*itemPtr;
    Elf_Segment	*segmentPtr;
    int		offset;

    /* 
     * Write the ELF header.
     */

    filePtr->ehdr.e_phnum = filePtr->numSegments;
    filePtr->ehdr.e_shnum = filePtr->numSections;
    len = filePtr->ehdr.e_ehsize;
#ifdef DEBUG
    printf("Writing %d bytes of ELF hdr at offset %ld\n", len, ftell(f));
#endif /* DEBUG */
    n = fwrite(&filePtr->ehdr, len, 1, f);
    assume(n == 1);
    /* 
     * Write the program headers.
     */
    offset = ftell(f);
    if (offset != filePtr->ehdr.e_phoff) {
	fprintf(stderr, 
	    "Incorrect program header offset in ELF header (%d != %d).\n",
	    filePtr->ehdr.e_phoff, offset);
	abort();
    }
    i = 0;
    LIST_FORALL(&filePtr->segmentList, itemPtr) {
	segmentPtr = (Elf_Segment *) itemPtr; 
	len = filePtr->ehdr.e_phentsize;
	printf("Writing %d bytes of program hdr %d at offset %ld\n", 
	    len, i, ftell(f));
	n = fwrite(&segmentPtr->phdr, len, 1, f);
	assume(n == 1);
	i++;
    }
    /*
     * Write the sections.
     */
    i = 0;
    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	sectionPtr = (Elf_Section *) itemPtr; 
	if ((sectionPtr->shdr.sh_type != SHT_NOBITS) && 
	    (sectionPtr->shdr.sh_type != SHT_NULL) && 
	    (sectionPtr->shdr.sh_type != SHT_LOUSER) && 
	    (sectionPtr->shdr.sh_size != 0)) {
	    n = fseek(f, sectionPtr->shdr.sh_offset, SEEK_SET);
	    assume(n == 0);
	    len = sectionPtr->shdr.sh_size;
#ifdef DEBUG
	    printf("Writing %d bytes of section %d at offset %ld\n", 
		len, i, ftell(f));
#endif /* DEBUG */
	    n = fwrite(sectionPtr->section, len, 1, f);
	    assume(n == 1);
	}
	i++;
    }
    /*
     * Write the section headers. 
     */
    offset = ftell(f);
    if (offset != filePtr->ehdr.e_shoff) {
	fprintf(stderr, 
	    "Incorrect section header offset in ELF header (%d != %d).\n",
	    filePtr->ehdr.e_shoff, offset);
	abort();
    }
    i = 0;
    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	sectionPtr = (Elf_Section *) itemPtr; 
	len = filePtr->ehdr.e_shentsize;
#ifdef DEBUG
	printf("Writing %d bytes of section hdr %d at offset %ld\n", 
		len, i, ftell(f));
#endif /* DEBUG */
	n = fwrite(&sectionPtr->shdr, filePtr->ehdr.e_shentsize, 1, f);
	assume(n = 1);
	i++;
    }
    return SUCCESS;
}

Elf_File *
Elf_FileNew(void)
{
    Elf32_Ehdr	*ehdrPtr;
    Elf_File	*filePtr;

    filePtr = xmalloc(sizeof(Elf_File));
    memset(filePtr, 0, sizeof(Elf_File));
    List_InitElement((List_Links *) filePtr)
    List_Init(&filePtr->segmentList);
    List_Init(&filePtr->sectionList);
    ehdrPtr = &filePtr->ehdr;
    ehdrPtr->e_ident[EI_MAG0] = ELFMAG0;
    ehdrPtr->e_ident[EI_MAG1] = ELFMAG1;
    ehdrPtr->e_ident[EI_MAG2] = ELFMAG2;
    ehdrPtr->e_ident[EI_MAG3] = ELFMAG3;
    ehdrPtr->e_ident[EI_CLASS] = ELFCLASS32;
    ehdrPtr->e_ident[EI_DATA] = ELFDATA2LSB;
    ehdrPtr->e_ident[EI_VERSION] = EV_CURRENT;
    ehdrPtr->e_type = ET_EXEC;
    ehdrPtr->e_machine = EM_386;
    ehdrPtr->e_version = EV_CURRENT;
    ehdrPtr->e_phentsize = sizeof(Elf32_Phdr);
    ehdrPtr->e_shentsize = sizeof(Elf32_Shdr);
    ehdrPtr->e_ehsize = sizeof(Elf32_Ehdr);
    ehdrPtr->e_phoff = sizeof(Elf32_Ehdr);
    return filePtr;
}

Elf_Section *
Elf_SectionNew(void)
{
    Elf_Section	 *sectionPtr;

    sectionPtr = xmalloc(sizeof(Elf_Section));
    memset(sectionPtr, 0, sizeof(Elf_Section));
    List_InitElement((List_Links *) sectionPtr);
    return sectionPtr;
}


Elf_Segment *
Elf_SegmentNew(void)
{
    Elf_Segment	 *segmentPtr;

    segmentPtr = xmalloc(sizeof(Elf_Segment));
    memset(segmentPtr, 0, sizeof(Elf_Segment));
    List_InitElement((List_Links *) segmentPtr);
    return segmentPtr;
}

Status
Elf_FileDump(
    FILE	*f,
    Elf_File	*filePtr)
{
    Elf_Section	*sectionPtr;
    Elf_Segment *segmentPtr;
    int		i;
    List_Links	*itemPtr;

    fprintf(f, "File: %s\n", filePtr->name);
    fprintf(f, "ELF header:\n");
    fprintf(f, "# segments: %d\n", filePtr->ehdr.e_phnum);
    fprintf(f, "# sections: %d\n", filePtr->ehdr.e_shnum);
    fprintf(f, "Segments:\n");
    fprintf(f, "Type\tOffset\tVA\t\tPA\tFS\t\tMS\n");
    LIST_FORALL(&filePtr->segmentList, itemPtr) {
	segmentPtr = (Elf_Segment *) itemPtr;
	fprintf(f, "%d\t0x%x\t0x%x\t0x%x\t0x%x\t0x%x\n", 
	    segmentPtr->phdr.p_type, segmentPtr->phdr.p_offset,
	    segmentPtr->phdr.p_vaddr, segmentPtr->phdr.p_paddr,
	    segmentPtr->phdr.p_filesz, segmentPtr->phdr.p_memsz);
    }
    fprintf(f, "Sections:\n");
    fprintf(f, "%-3s %-40s %-10s %-10s %-10s %-10s\n", "#", "Name", "Type", 
	"Addr", "Offset", "Size");
    i = 0;
    LIST_FORALL(&filePtr->sectionList, itemPtr) {
	sectionPtr = (Elf_Section *) itemPtr;
	fprintf(f, "%3d %-40s 0x%08x 0x%08x 0x%08x 0x%08x\n", 
	    i, sectionPtr->name, sectionPtr->shdr.sh_type, 
	    sectionPtr->shdr.sh_addr, sectionPtr->shdr.sh_offset, 
	    sectionPtr->shdr.sh_size);
	i++;
    }
    return SUCCESS;
}


