/* 
 * status.c --
 *
 *	Turns a status value into a string.
 *
 * @COPYRIGHT@
 */

#ifndef lint
static char rcsid[] = "$Header: /cvs/cvs/slinky/slinky/src2/libs/status/status.c,v 1.1 2004/01/16 20:00:07 jhh Exp $";
#endif /* not lint */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <status.h>
#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <assert.h>
#include <errno.h>

typedef struct {
    char	*string;	// String associated with msg
    int		errnoValue;	// Corresponding errno
} MsgInfo;

typedef struct {
    char	*name;		// Module name
    MsgInfo	*info;		// Error information
    int		size;		// Number of messages
} StatusMsgs;


static MsgInfo elfStatusMsgs[] = {
    {"bad magic in header", 			EINVAL},
    {"bad class in header",			EINVAL},
    {"bad data in header",			EINVAL},
    {"bad version in header",			EINVAL},
    {"bad ELF header size",			EINVAL},
    {"bad section header size",			EINVAL},
    {"bad program header size",			EINVAL},
    {"I/O error",				EINVAL},
    {"a parameter is out of bounds",		EINVAL},
};

static MsgInfo statusStatusMsgs[] = {
    {"invalid status code",			EINVAL},
};

#define NUM_CODES(array) (sizeof(array) / sizeof(MsgInfo))
static StatusMsgs statusMsgs[] = {
    {"Errno",	NULL, 0},
    {"ELF",	elfStatusMsgs, NUM_CODES(elfStatusMsgs)},
    {"Status",	statusStatusMsgs, NUM_CODES(statusStatusMsgs)},
};

static Status	ValidStatus(Status status, char **errorMsg);

/*
 *----------------------------------------------------------------------
 *
 * Status_Msg --
 *
 *	Returns the string corresponding to the given status.
 *
 * Results:
 *	Statically allocated string.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

char *
Status_Msg(status)
    Status	status;
{
    unsigned int	module = STATUS_MODULE(status);
    unsigned int	msg = STATUS_MSGNUM(status);
    Status		code;
    char		*error;

    code = ValidStatus(status, &error);
    if (code != SUCCESS) {
	return error;
    }
    /* 
     * Module 0 corresponds to errno values. Linux uses -errno a lot, which
     * corresponds to module 0xffff.
     */
    if (module == 0) {
	return strerror(status);
    } if (module == 0xffff) {
	return strerror(-status);
    } else if (status == FAILURE) {
	return "generic failure";
    } else {
	return statusMsgs[module].info[msg].string;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Status_ToErrno --
 *
 *	Converts a Swarm status into the most appropriate errno.
 *
 * Results:
 *	STATUS_BADSTATUS if status is invalid, SUCCESS otherwise
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Status
Status_ToErrno(status, errnoValue)
    Status		status;		// Status to be converted.
    int			*errnoValue;	// Returned errno value.
{

    Status		code;
    unsigned int	module = STATUS_MODULE(status);
    unsigned int	msg = STATUS_MSGNUM(status);
    char		*error;

    code = ValidStatus(status, &error);
    if (code != SUCCESS) {
	return STATUS_BADSTATUS;
    }
    /* 
     * Module 0 corresponds to errno values. Linux uses -errno a lot,
     * which corresponds to module 0xffff
     */
    if (module == 0) {
	*errnoValue = status;
    } else if (module == 0xffff) {
	*errnoValue = -status;
    } else if (status == FAILURE) {
	*errnoValue = EINVAL;
    } else {
	*errnoValue = statusMsgs[module].info[msg].errnoValue;
    }
    return SUCCESS;
}

/*
 *----------------------------------------------------------------------
 *
 * ValidStatus --
 *
 *	Checks if the given status is valid.
 *
 * Results:
 *	SUCCESS or FAILURE
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Status
ValidStatus(status, errorMsg)
    Status	status;		// Status to be checked for validity
    char	**errorMsg;	// Error message
{
    unsigned int	module = STATUS_MODULE(status);
    unsigned int	msg = STATUS_MSGNUM(status);
    static char		buffer[100];

    /*
     * Make sure a Status is 32 bits.
     */
    assert(sizeof(Status) == 4);
    if (status == FAILURE) {
	return SUCCESS;
    }
    /* 
     * Module 0 corresponds to errno values.
     */
    if (module == 0) {
#if !(defined(LINUX) && defined(__KERNEL__))
	*errorMsg = strerror(msg);
	if (*errorMsg == NULL) { 
	    return FAILURE;
	}
#endif /* !(defined(LINUX) && defined(__KERNEL__)) */
    } else if (module == 0xffff) {
#if !(defined(LINUX) && defined(__KERNEL__))
	msg = STATUS_MSGNUM(-status);
	*errorMsg = strerror(msg);
	if (*errorMsg == NULL) { 
	    return FAILURE;
	}
#endif /* !(defined(LINUX) && defined(__KERNEL__)) */
    } else if (module >= (sizeof(statusMsgs) / sizeof(StatusMsgs))) {
	(void) sprintf(buffer, "Invalid module 0x%x in status 0x%x",
		module, status);
	*errorMsg = buffer;
	return FAILURE;
    } else if (msg >= statusMsgs[module].size) {
	(void) sprintf(buffer, "Invalid msg 0x%x for %s module in status 0x%x",
		msg, statusMsgs[module].name, status);
	*errorMsg = buffer;
	return FAILURE;
    }
    return SUCCESS;
}

